/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Jasper van de Gronde (win32 port)
 */

#include <string.h>
#include <stdio.h>
#include "rdf-int.h"
#include "indexPool.h"
#include "../utils/utils.h"
#include "../utils/hash.h"

extern int gAssertionCount = 0;
char *getFileCode(RDF_Parser rdfp);

extern FILE *rdfpGetFile(RDF_Parser rdfp);

/* JvdG */
typedef struct _RDF_Serializer RDF_Serializer;
struct _RDF_Serializer {
  void* (*create)(const char* url, int raw);
  RDF_Result (*serializeTriple)(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data);
  void (*destroy)(void* serializer_data);
};

static RDF_Serializer* RDF_GetSerializer(RDF_SerializationType type);

static void* NTriplesSerializer_create(const char*, int raw);
static RDF_Result NTriplesSerializer_serializeTriple(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data);
static void NTriplesSerializer_destroy(void* serializer_data);
static RDF_Serializer NTriplesSerializer = {NTriplesSerializer_create, NTriplesSerializer_serializeTriple, NTriplesSerializer_destroy};

static void* TurtleSerializer_create(const char*, int raw);
static RDF_Result TurtleSerializer_serializeTriple(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data);
static void TurtleSerializer_destroy(void* serializer_data);
static RDF_Serializer TurtleSerializer = {TurtleSerializer_create, TurtleSerializer_serializeTriple, TurtleSerializer_destroy};

static RDF_Result CSVSerializer_serializeTriple(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data);
static RDF_Serializer CSVSerializer = {NTriplesSerializer_create, CSVSerializer_serializeTriple, NTriplesSerializer_destroy};


void initIndexPool();
void deinitIndexPool();

RDF_Result 
RDF_Initialize() {
  initIndexPool();
  DB_Initialize();
  return RDF_OK;
}

RDF_Result 
RDF_Deinitialize() {
  deinitIndexPool () ;
  DB_Deinitialize();
  return RDF_OK;
}


char* 
getIndexString(RDF_Resource res) {
  return (char*) res;
}

char* 
dbName (RDF_DB db) {
  return (char*) db;
}

/* JvdG Fixed this for when item_type == RDF_INT_TYPE.
        Also removed all strcat's in favour of memcpy to a direct address. */
Index
makeIndex (const void* item_value, RDF_ValueType item_type, const char* arc) {
  size_t resLen ;
  size_t arcLen; 
  Index ind ;

  if ((item_type == RDF_RESOURCE_TYPE) || (item_type == RDF_STRING_TYPE)) {
    const char* res = (const char*) item_value;
    char* data;
    resLen = strlen(res);
    arcLen = strlen(arc);
    ind = getIndexForLen(resLen + 1 + arcLen + 1);
    data = ind->data;
    memcpy(data, res, resLen); data += resLen;
    *data = '|'; ++data;
    memcpy(data, arc, arcLen);
    ind->len = resLen + 1 + arcLen;
    return ind;
  } else if (item_type == RDF_INT_TYPE) {
    char res[12];
    char* data;
    snprintf(res, 12, "%i", (int) item_value);
    resLen = strlen(res);
    arcLen = strlen(arc);
    ind = getIndexForLen(resLen + 1 + arcLen + 1);
    data = ind->data;
    memcpy(data, res, resLen); data += resLen;
    *data = '|'; ++data;
    memcpy(data, arc, arcLen);
    ind->len = resLen + 1 + arcLen;
    return ind;
  } else {
    return 0;
  }
}

/* JvdG Removed all strcat's in favour of memcpy to a direct address.
        Replaced \n by \xA to make sure the right character code is used. */
Index
makeValue (const void* item_value, RDF_ValueType item_type, const char* sourceCode) {
  size_t resLen ;
  size_t sl = strlen(sourceCode);
  Index ind ;

  if ((item_type == RDF_RESOURCE_TYPE) || (item_type == RDF_STRING_TYPE)) {
    const char* res = (const char*) item_value;
    char* data;
    resLen = strlen(res);
    ind = getIndexForLen(2 + sl + 1 + resLen);
    ind->data[0] = (char)(item_type == RDF_RESOURCE_TYPE ? '1' : '3');
    ind->data[1] = '\t';
    data = ind->data+2;
    memcpy(data, sourceCode, sl); data += sl;
    *data = '\xA'; ++data;
    memcpy(data, res, resLen); data += resLen;
    ind->len = data - ind->data;
    return ind;
  } else if (item_type == RDF_INT_TYPE) {
    char res[32];
    snprintf(res, 32, "%i\t%s" "\xA" "%i", RDF_INT_TYPE, sourceCode, (int) item_value);
    resLen = strlen(res);
    ind = getIndexForLen(resLen);
    memcpy(ind->data, res, resLen);
    ind->len = resLen;
    return ind;
  } else {
    return 0;
  }
}
    

int
addRemoveTriple (RDF_DB db,  RDF_Resource source, RDF_Resource arc, 
                 const void* target_value, RDF_ValueType target_type,
                 RDF_Operation op, const char* sourceCode) {

  const char*          Src = source;
  const char*     realSrc  = getIndexString(source);
  char*           realArc  = getIndexString(arc);
  const void*   realTarget = (target_type == RDF_RESOURCE_TYPE ? getIndexString((RDF_Resource) target_value) : target_value);
  
  Index forwardIndex = makeIndex(realSrc, RDF_RESOURCE_TYPE, realArc);
  Index reverseIndex = makeIndex(realTarget, target_type, realArc);
  Index forwardValue = makeValue(realTarget, target_type, sourceCode);
  Index reverseValue = makeValue(realSrc, RDF_RESOURCE_TYPE, sourceCode);
  Index arcValue     = makeValue(Src, RDF_RESOURCE_TYPE, sourceCode);
  int err = 0;

  if (op == RDF_ASSERT) {
    /* JvdG Changed this to mirror the behaviour when op == RDF_UNASSERT */
    err = DB_AddForwardIndex(db, forwardIndex->data, forwardValue->data, 1);
    if (!err && reverseIndex) err = DB_AddReverseIndex(db, reverseIndex->data,
                                                         reverseValue->data, 0);
    if (!err) err = DB_AddArcIndex(db, realArc, arcValue->data, 0);
    /* JvdG This only stores each arc/source pair once in the arcindex, which is lovely from a space point of view, but not very safe (it gets deleted even when only one triple with that arc/source pair is deleted).
            A future optimization may use something like reference counting or different behaviour for load and insert queries (although it's questionable whether that would be wanted). */
    /*
    int err1 = DB_AddForwardIndex(db, forwardIndex->data, forwardValue->data, 1);
    if ((err1 > -1) && (err1 != 2)) {
      if (reverseIndex) 
        err = DB_AddReverseIndex(db, reverseIndex->data, reverseValue->data, 0);
      if (!err && (err1 != 1)) err = DB_AddArcIndex(db, realArc, arcValue->data, 0);
    }
    else {
      err = err1;
    }*/
  } else if (op == RDF_UNASSERT) {
    err = DB_RemoveForwardIndex(db, forwardIndex->data, forwardValue->data);
    if (!err && reverseIndex) err = DB_RemoveReverseIndex(db, reverseIndex->data, 
                                                          reverseValue->data);
    if (!err) err = DB_RemoveArcIndex(db, realArc, arcValue->data);
  }

  freeIndex(forwardIndex);
  freeIndex(reverseIndex);
  freeIndex(forwardValue);
  freeIndex(reverseValue);
  freeIndex(arcValue);

  return err;
  
}

const RDF_Resource 
RDF_GetResource (const char* uri) {
  return (const RDF_Resource)uri;
}

RDF_Result
RDF_GetDB (char* name, int createp, RDF_DB *db) {
  int err = DB_OpenDatabase (name, createp); /* JvdG Changed this to not always create the db. */
  if (!err) {
    *db = name;
    return 0;
  } else {
    return err;
  }
}


/* JvdG
 * Made this use < and >.
 * Updated the output format to the new notation (src arc dst).
 */
void
printAssertion (VerbosityLevel level, char* prefix, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) {
  PrintMsg(level, "%s (<%s> <%s> ", prefix, (char*)source, (char*)arc);
  if (target_type == RDF_RESOURCE_TYPE) {
    PrintMsg(level, "<%s>)", (char*) target_value);
  }
  else if (target_type == RDF_STRING_TYPE) {
    PrintMsg(level, "\'%s\')", (char*) target_value);
  }
  else if (target_type == RDF_INT_TYPE) {
    PrintMsg(level, "%i)", (int) target_value);
  }
  else if (target_type == 0) {
    PrintMsg(level, "%s)", (char*) 0);
  }
  else {
    PrintMsg(level, "UNKNOWN_TYPE)");
  }
}


/* JvdG
 * Made this detect whether or not a triple is added.
 */
RDF_Result 
RDF_Assert (RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) {
  int err;
  printAssertion(DEBUG_INFO, "\nAsserting:", source, arc, target_value, target_type);  
  err = addRemoveTriple(db, source, arc, target_value, target_type, RDF_ASSERT, "0");
  if ( !err ) gAssertionCount++;
  if ( err == 2 ) err = 0;
  return err;
}

/* JvdG
 * Made this detect whether or not a triple is added.
 * It now sets err to 0 if the assertion already existed.
 */
RDF_Result 
RDF_LoadFileAssert (RDF_Parser rdfp, RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) {
  int err;
  printAssertion(DEBUG_INFO, "\nAsserting:", source, arc, target_value, target_type);  
  err = addRemoveTriple(db, source, arc, target_value, target_type, RDF_ASSERT, getFileCode(rdfp));
  if (!err ) {
    gAssertionCount++;
    if ( (rdfpGetFile(rdfp)) ) {
      FILE *file = rdfpGetFile(rdfp);
      fprintf(file, "%s\t%s\t", arc, source);
      if (target_type == RDF_INT_TYPE) {
        fprintf(file, "%i\n", (int)target_value);
      } else if (target_type == RDF_STRING_TYPE) {
        fprintf(file, "'%s'\n", (char*)target_value);
      } else if (target_type == RDF_RESOURCE_TYPE) {
        fprintf(file, "<%s>\n", (char*)target_value);
      }
    }
  }
  if ( err == 2 ) err = 0;
  return 0;
}

/* JvdG Let this increment gAssertionCount */
RDF_Result 
RDF_Unassert (RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) {
  int err;
  printAssertion(DEBUG_INFO, "\nUnasserting:", source, arc, target_value, target_type);  
  err = addRemoveTriple(db, source, arc, target_value, target_type, RDF_UNASSERT, "0");
  if ( !err ) gAssertionCount++;
  if ( err == 2 ) err = 0;
  return err;
}

/* JvdG
 * Added this function as counterpart for RDF_LoadFileAssert.
 * Let this increment gAssertionCount
 */
RDF_Result 
RDF_UnloadFileAssert (const char* sourceCode, RDF_DB db, RDF_Resource source, RDF_Resource arc, const void* target_value, RDF_ValueType target_type) {
  int err;
  printAssertion(DEBUG_INFO, "\nUnasserting:", source, arc, target_value, target_type);  
  err = addRemoveTriple(db, source, arc, target_value, target_type, RDF_UNASSERT, sourceCode);
  if ( !err ) gAssertionCount++;
  if ( err == 2 ) err = 0;
  return err;
}

/* JvdG Made this accept an RDF_Value and use printAssertion */
RDF_Result 
RDF_HasAssertion(RDF_DB db, const RDF_Resource source, const RDF_Resource arc, const RDF_Value target) {
  const char* realSrc = getIndexString(source);
  char*     realArc   = getIndexString(arc);
  Index forwardIndex  = makeIndex(realSrc, RDF_RESOURCE_TYPE, realArc);
  char* dbn           = dbName(db);
  int err = 0;
  DB_Cursor dbc = 0;

  if (0 != (err = DB_GetForwardCursor(dbn, forwardIndex->data, &dbc))) {
    return err;
  }

  printAssertion(DEBUG_INFO, "\nHas Assertion ", source, arc, target.value, target.type);

  while (0 == DB_CursorAdvance(dbc)) {
    int vtype ;
    void *value;
    DB_CursorCurrentValue(dbc , &value);
    DB_CursorCurrentValueType(dbc, &vtype);

    if (vtype == target.type) {
      if ((target.type == RDF_RESOURCE_TYPE) || (target.type == RDF_STRING_TYPE)) {
        if (streq(value, (char*)target.value)) return 1;
      } else if (target.type == RDF_INT_TYPE) {
        if ((int)value == (int) target.value) return 1;
      }
    }
  }
  
  DB_ReleaseCursor(dbc);
  return 0;
}

/* JvdG Made this use an RDF_Value */
RDF_Result
RDF_GetTargets (RDF_DB db, RDF_Resource source, RDF_Resource arc, RDF_Cursor *cursor) {
  const char* realSrc = getIndexString(source);
  char*     realArc   = getIndexString(arc);
  Index forwardIndex  = makeIndex(realSrc, RDF_RESOURCE_TYPE, realArc);
  char* dbn           = dbName(db);
  DB_Cursor dbc = 0;
  int       err = 0;
  RDF_Cursor c;

  if  (0 != (err  = DB_GetForwardCursor(dbn, forwardIndex->data, &dbc))) {
    return err;
  }

  PrintMsg(DEBUG_INFO, "\nGet Targets source=%s, arc=%s", source, arc);

  if (dbc) {    
    c = getNewCursor();
    c->dbCursor = dbc;
    c->db = db;
    c->res = source;
    c->arc = arc;
    c->inversep = 0;
    c->index = forwardIndex;
  }

  *cursor = c;
  return 0;
}

/* JvdG Made this use an RDF_Value */
RDF_Result
RDF_GetArcsOut (RDF_DB db, RDF_Resource source,  RDF_Cursor *cursor) {
  const char* realSrc = getIndexString(source);
  char* dbn           = dbName(db);
  Index ind           = makeIndex(realSrc, RDF_RESOURCE_TYPE, "");
  DB_Cursor dbc = 0;
  int       err = 0;
  RDF_Cursor c;

  if  (0 != (err  = DB_GetArcsOutCursor(dbn, ind->data, &dbc))) {
    return err;
  }

  PrintMsg(DEBUG_INFO, "\nGet Arcs Out =%s",  source);

  if (dbc) {    
    c = getNewCursor();
    c->dbCursor = dbc;
    c->db = db;
    c->res = source;
    c->index = ind;
  }

  *cursor = c;
  return 0;
}

/* JvdG Made this accept an RDF_Value.
   I'm not sure about whether it is necessary to support other types than RDF_RESOURCE_TYPE
   here, but it probably won't hurt either. */
RDF_Result
RDF_GetArcsIn (RDF_DB db, RDF_Value target,  RDF_Cursor *cursor) {
  const void* realTrg = (target.type == RDF_RESOURCE_TYPE ? getIndexString((RDF_Resource) target.value) : target.value);
  char* dbn           = dbName(db);
  Index ind           = makeIndex(realTrg, target.type, "");
  DB_Cursor dbc = 0;
  int       err = 0;
  RDF_Cursor c;

  if  (0 != (err  = DB_GetArcsInCursor(dbn, ind->data, &dbc))) {
    return err;
  }

  switch(target.type) {
  case RDF_RESOURCE_TYPE:
    PrintMsg(DEBUG_INFO, "\nGet Arcs In = <%s>", target.value);
    break;
  case RDF_STRING_TYPE:
    PrintMsg(DEBUG_INFO, "\nGet Arcs In = '%s'", target.value);
    break;
  case RDF_INT_TYPE:
    PrintMsg(DEBUG_INFO, "\nGet Arcs In = %i", target.value);
    break;
  }

  if (dbc) {    
    c = getNewCursor();
    c->dbCursor = dbc;
    c->db = db;
    c->res = target.value;
    c->index = ind;
  }

  *cursor = c;
  return 0;
}
 

RDF_Result 
RDF_GetArcSources (RDF_DB db, RDF_Resource arc, RDF_Cursor *cursor) {
  char* realArc = getIndexString(arc);
  char* dbn     = dbName(db);
  DB_Cursor dbc = 0;
  int       err = 0;
  RDF_Cursor c;

  if  (0 != (err  = DB_GetArcCursor(dbn, realArc, &dbc))) {
    return err;
  }

  PrintMsg(DEBUG_INFO, "\nGet Arc sources arc=%s",  realArc);

  if (dbc) {    
    c = getNewCursor();
    c->dbCursor = dbc;
    c->db = db;
    c->arc = arc;
    c->inversep = 0;
  }

  *cursor = c;
  return 0;
}

/* JvdG Made this accept an RDF_Value */
RDF_Result
RDF_GetSources(RDF_DB db, RDF_Value target, RDF_Resource arc, RDF_Cursor *cursor) {
  const void* realTrg = (target.type == RDF_RESOURCE_TYPE ? getIndexString((RDF_Resource) target.value) : target.value);
  char*     realArc   = getIndexString(arc);
  Index reverseIndex  = makeIndex(realTrg, target.type, realArc); 
  char* dbn           = dbName(db);
  DB_Cursor dbc  = 0 ;
  int       err = 0;
  RDF_Cursor c;

  if  (0 != (err  = DB_GetReverseCursor(dbn, reverseIndex->data, &dbc))) {
    return err;
  }

  switch(target.type) {
  case RDF_RESOURCE_TYPE:
    PrintMsg(DEBUG_INFO, "\nGet Sources target=<%s>, arc=%s", target.value, arc);
    break;
  case RDF_STRING_TYPE:
    PrintMsg(DEBUG_INFO, "\nGet Sources target='%s', arc=%s", target.value, arc);
    break;
  case RDF_INT_TYPE:
    PrintMsg(DEBUG_INFO, "\nGet Sources target=%i, arc=%s", target.value, arc);
    break;
  }
  
  if (dbc) {    
    c = getNewCursor();
    c->dbCursor = dbc;
    c->db = db;
    c->res = target.value;
    c->arc = arc;
    c->inversep = 1;
    c->index = reverseIndex;
  }

  *cursor = c;
  return 0;
}

/* JvdG */
/** Currently this assumes all values are null (it finds all triples). */
RDF_Result
RDF_FindStatements(RDF_DB db, RDF_Resource source, RDF_Resource arc, RDF_Value target,  RDF_Cursor *cursor) {
  char* dbn     = dbName(db);
  DB_Cursor dbc = 0;
  int       err = 0;
  RDF_Cursor c;

  if  (0 != (err  = DB_GetEverythingCursor(dbn, &dbc))) {
    return err;
  }

  printAssertion(DEBUG_INFO, "\nFind Statements: ", source, arc, target.value, target.type);

  if (dbc) {    
    c = getNewCursor();
    c->dbCursor = dbc;
    c->db = db;
  }

  *cursor = c;
  return 0;
}

RDF_Result
RDF_Advance (RDF_Cursor c) {
  if (!c) {
    return RDF_NO_MORE_VALUES;
  } else {
    return (RDF_Result) DB_CursorAdvance(c->dbCursor);
  }
}

RDF_Result 
RDF_CursorCurrentValue (RDF_Cursor c, void** value) {
  if (!c) {
    return RDF_NO_MORE_VALUES;
  } else {
    return DB_CursorCurrentValue(c->dbCursor, value);
  }
}

RDF_Result 
RDF_CursorCurrentValueSize (RDF_Cursor c, int *sz) {
  if (!c) {
    return RDF_NO_MORE_VALUES;
  } else {
    return DB_CursorCurrentValueSize(c->dbCursor, sz);
  }
}
RDF_Result 
RDF_CursorCurrentValueType (RDF_Cursor c, RDF_ValueType *type) {
  if (!c) {
    return RDF_NO_MORE_VALUES;
  } else {
    return DB_CursorCurrentValueType(c->dbCursor, (int*)type);
  }
}

/* JvdG */
RDF_Result 
RDF_CursorCurrentSource (RDF_Cursor c, void** value) {
  if (!c) {
    return RDF_NO_MORE_VALUES;
  } else {
    return DB_CursorCurrentSource(c->dbCursor, value);
  }
}

/* JvdG */
RDF_Result 
RDF_CursorCurrentArc (RDF_Cursor c, void** value) {
  if (!c) {
    return RDF_NO_MORE_VALUES;
  } else {
    return DB_CursorCurrentArc(c->dbCursor, value);
  }
}

RDF_Result
RDF_DisposeCursor(RDF_Cursor cursor) {
  DB_ReleaseCursor(cursor->dbCursor); 
  if (cursor->index) freeIndex(cursor->index);
  cursor->index = 0;
  freeCursor(cursor);
  return 0;
}

/* JvdG */
RDF_Result 
RDF_ExportDB(RDF_DB db, const char* url, RDF_SerializationType type, int raw) {
  RDF_Cursor c = 0;
  const RDF_Value RDFNull = {0,0};
  RDF_Result err;
  RDF_Serializer* serializer;
  void* serializer_data;

  serializer = RDF_GetSerializer(type);
  if ( !serializer ) return RDF_UNKNOWN_FILE_FORMAT_ERROR;
  serializer_data = serializer->create(url, raw);
  if ( !serializer_data ) return RDF_ERROR;
  
  err = RDF_FindStatements(db, 0, 0, RDFNull, &c);

  if ( err != RDF_OK ) return err;
  if ( !c ) return RDF_COULD_NOT_ACCESS_DATA;

  while (0 == RDF_Advance(c)) {
    RDF_Value source;
    RDF_Value arc;
    RDF_Value value;
    int len;

    RDF_CursorCurrentSource(c, &(source.value));
    source.type = RDF_RESOURCE_TYPE;
    RDF_CursorCurrentArc(c, &(arc.value));
    arc.type = RDF_RESOURCE_TYPE;
    RDF_CursorCurrentValue(c, &(value.value));
    RDF_CursorCurrentValueSize(c, &len);
    RDF_CursorCurrentValueType(c, &(value.type));
    if (value.type == RDF_RESOURCE_TYPE || value.type == RDF_STRING_TYPE) {
      ((char*)value.value)[len] = 0; 
    }

    err = serializer->serializeTriple(source, arc, value, serializer_data);
    if ( err != RDF_OK ) break;

    ++gAssertionCount;
  }

  RDF_DisposeCursor(c);
  serializer->destroy(serializer_data);

  return err;
}

/* JvdG */
RDF_Serializer* RDF_GetSerializer(RDF_SerializationType type) {
  switch(type) {
  case NTRIPLES:
    return &NTriplesSerializer;
  case TURTLE:
    return &TurtleSerializer;
  case CSV:
    return &CSVSerializer;
  default:
    return 0;
  }
}

/***** N-Triples Serializer *****/

/* JvdG */
typedef struct _BlankNodeData {
  HashTable ht;
  char lastNodeID[20];
  int raw;
} BlankNodeData;

/* JvdG */
typedef struct _NTriplesSerializerUserData {
  FILE* file;
  char* buffer;
  size_t buffer_size;
  BlankNodeData bndata;
} NTriplesSerializerUserData;

/* JvdG */
void incrementNodeID(char* nodeID) {
  do {
    if ( *nodeID ) {
      *nodeID = ((*nodeID-'a'+1) % 26) + 'a';
    }
    else {
      *nodeID++ = 'a';
      *nodeID = 0;
    }
  } while(*nodeID++ == 'a');
}

/* JvdG */
static void* NTriplesSerializer_create(const char* url, int raw) {
  NTriplesSerializerUserData* user_data = 0;
  FILE* file = 0;
  if ( startsWith(url, "file:") ) {
    url += 5;
    if ( startsWith(url, "///") ) url += 2;
#ifdef WIN32
    if ( *url == '/' ) ++url;
#endif
    file = fopen(url, "wt");
  }
  if ( !file ) {
    return 0;
  }
  user_data = (NTriplesSerializerUserData*)getMem(sizeof(NTriplesSerializerUserData));
  if ( user_data ) {
    user_data->file = file;
    user_data->bndata.raw = raw;
    user_data->bndata.ht = NewHashTable(1024);
    user_data->bndata.lastNodeID[0] = 0;
    if ( !user_data->bndata.ht ) {
      fclose(user_data->file);
      freeMem(user_data);
      return 0;
    }
  }
  return user_data;
}

/* JvdG */
static RDF_Result N3Serializer_serializeValue(RDF_Value v, FILE* file, char** buffer, size_t* buffer_size, BlankNodeData* bndata) {
  switch(v.type) {
  case RDF_RESOURCE_TYPE:
    if ( *((char*)v.value) == '_' ) {
      if ( bndata->raw ) {
        fprintf(file, "<BN%s>", (char*)v.value+2);
      }
      else {
        char* nodeID = HashLookup(bndata->ht, v.value);
        if ( !nodeID ) {
          incrementNodeID(bndata->lastNodeID);
          nodeID = strdup(bndata->lastNodeID);
          HashAdd(bndata->ht, v.value, nodeID);
        }
        fprintf(file, "_:%s", nodeID);
      }
    }
    else {
      fprintf(file, "<%s>", v.value);
    }
    break;
  case RDF_STRING_TYPE:
    if ( !reserve(buffer, buffer_size, 1+2*strlen(v.value)+1+1) ) return RDF_ERROR;
    outputEscapedString(v.value, *buffer);
    fprintf(file, "%s", *buffer);
    break;
  case RDF_INT_TYPE:
    fprintf(file, "%i", v.value);
    break;
  case RDF_FLOAT_TYPE:
    {
      double dval = *((float*)(&v.value));
      fprintf(file, "%f", dval);
    }
    break;
  case RDF_DATE_TYPE:
  case RDF_BLOB_TYPE:
  default:
    return RDF_ERROR;
  }
  return RDF_OK;
}

/* JvdG */
static RDF_Result NTriplesSerializer_serializeTriple(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data) {
  NTriplesSerializerUserData* user_data = serializer_data;
  FILE* file = user_data->file;
  RDF_Result err;
  err = N3Serializer_serializeValue(source, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
  if ( err != RDF_OK ) return err;
  fputs(" ", file);
  err = N3Serializer_serializeValue(arc, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
  if ( err != RDF_OK ) return err;
  fputs(" ", file);
  err = N3Serializer_serializeValue(target, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
  if ( err != RDF_OK ) return err;
  fputs(" .\n", file);
  return RDF_OK;
}

/* JvdG */
static void NTriplesSerializer_destroy(void* serializer_data) {
  NTriplesSerializerUserData* user_data = serializer_data;
  FILE* file = user_data->file;
  if ( file ) {
    fclose(file);
  }
  if ( user_data->buffer ) {
    free(user_data->buffer);
  }
  {
    int i;
    for(i = 0; i < user_data->bndata.ht->size; ++i) {
      HashEntry he = user_data->bndata.ht->buckets[i];
      while(he) {
        free(he->value);
        he = he->next;
      }
    }
    DeleteHashTable(user_data->bndata.ht);
  }
  freeMem(user_data);
}

/***** Turtle Serializer *****/

/* JvdG */
typedef struct _TurtleSerializerUserData {
  FILE* file;
  char* buffer;
  size_t buffer_size;
  char* current_source;
  size_t current_source_size;
  char* current_arc;
  size_t current_arc_size;
  BlankNodeData bndata;
} TurtleSerializerUserData;

/* JvdG */
static void* TurtleSerializer_create(const char* url, int raw) {
  TurtleSerializerUserData* user_data = 0;
  FILE* file = 0;
  if ( startsWith(url, "file:") ) {
    url += 5;
    if ( startsWith(url, "///") ) url += 2;
#ifdef WIN32
    if ( *url == '/' ) ++url;
#endif
    file = fopen(url, "wt");
  }
  if ( !file ) {
    return 0;
  }
  user_data = (TurtleSerializerUserData*)getMem(sizeof(TurtleSerializerUserData));
  if ( user_data ) {
    user_data->file = file;
    user_data->bndata.raw = raw;
    user_data->bndata.ht = NewHashTable(1024);
    user_data->bndata.lastNodeID[0] = 0;
    if ( !user_data->bndata.ht ) {
      fclose(user_data->file);
      freeMem(user_data);
      return 0;
    }
    user_data->current_source = malloc(1);
    user_data->current_arc = malloc(1);
    if ( !user_data->current_source || !user_data->current_arc ) {
      if ( user_data->current_source ) free(user_data->current_source);
      if ( user_data->current_arc ) free(user_data->current_arc);
      DeleteHashTable(user_data->bndata.ht);
      fclose(user_data->file);
      freeMem(user_data);
      return 0;
    }
    *user_data->current_source = 0;
    *user_data->current_arc = 0;
    user_data->current_source_size = 1;
    user_data->current_arc_size = 1;
  }
  return user_data;
}

/* JvdG */
static RDF_Result TurtleSerializer_serializeTriple(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data) {
  TurtleSerializerUserData* user_data = serializer_data;
  FILE* file = user_data->file;
  RDF_Result err;

  if ( 0 != strcmp(source.value, user_data->current_source) ) {
    if ( *user_data->current_source ) {
      fputs(" .\n", file);
    }
    err = N3Serializer_serializeValue(source, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
    if ( err != RDF_OK ) return err;
    fputs(" ", file);
    err = N3Serializer_serializeValue(arc, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
    if ( err != RDF_OK ) return err;
    fputs(" ", file);
    if ( !reserve(&user_data->current_source, &user_data->current_source_size, strlen(source.value)+1) ) return RDF_ERROR;
    if ( !reserve(&user_data->current_arc, &user_data->current_arc_size, strlen(arc.value)+1) ) return RDF_ERROR;
    strcpy(user_data->current_source, source.value);
    strcpy(user_data->current_arc, arc.value);
  }
  else if ( 0 != strcmp(arc.value, user_data->current_arc) ) {
    fputs(" ;\n\t", file);
    err = N3Serializer_serializeValue(arc, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
    if ( err != RDF_OK ) return err;
    fputs(" ", file);
    if ( !reserve(&user_data->current_arc, &user_data->current_arc_size, strlen(arc.value)+1) ) return RDF_ERROR;
    strcpy(user_data->current_arc, arc.value);
  }
  else {
    fputs(", ", file);
  }
  err = N3Serializer_serializeValue(target, file, &user_data->buffer, &user_data->buffer_size, &user_data->bndata);
  if ( err != RDF_OK ) return err;
  return RDF_OK;
}

/* JvdG */
static void TurtleSerializer_destroy(void* serializer_data) {
  TurtleSerializerUserData* user_data = serializer_data;
  FILE* file = user_data->file;
  if ( file ) {
    if ( *user_data->current_source ) {
      fputs(" .\n", file);
    }
    fclose(file);
  }
  if ( user_data->buffer ) free(user_data->buffer);
  if ( user_data->current_arc ) free(user_data->current_arc);
  if ( user_data->current_source ) free(user_data->current_source);
  {
    int i;
    for(i = 0; i < user_data->bndata.ht->size; ++i) {
      HashEntry he = user_data->bndata.ht->buckets[i];
      while(he) {
        free(he->value);
        he = he->next;
      }
    }
    DeleteHashTable(user_data->bndata.ht);
  }
  freeMem(user_data);
}

/***** CSV Serializer *****/

/* JvdG */
typedef NTriplesSerializerUserData CSVSerializerUserData;

/* JvdG */
static RDF_Result CSVSerializer_serializeValue(RDF_Value v, CSVSerializerUserData* user_data) {
  FILE* file = user_data->file;
  switch(v.type) {
  case RDF_RESOURCE_TYPE:
    if ( *((char*)v.value) == '_' ) {
      char* nodeID;
      if ( user_data->bndata.raw ) {
        if ( !reserve(&user_data->buffer, &user_data->buffer_size, 1+2*strlen(v.value)+1+1) ) return RDF_ERROR;
        outputCSVEscapedString(v.value, user_data->buffer);
        nodeID = user_data->buffer;
      }
      else {
        nodeID = HashLookup(user_data->bndata.ht, v.value);
        if ( !nodeID ) {
          incrementNodeID(user_data->bndata.lastNodeID);
          nodeID = strdup(user_data->bndata.lastNodeID);
          HashAdd(user_data->bndata.ht, v.value, nodeID);
        }
      }
      fprintf(file, "BLANKNODE,%s,", nodeID);
    }
    else {
      if ( !reserve(&user_data->buffer, &user_data->buffer_size, 1+2*strlen(v.value)+1+1) ) return RDF_ERROR;
      outputCSVEscapedString(v.value, user_data->buffer);
      fprintf(file, "URI,%s,", user_data->buffer);
    }
    break;
  case RDF_STRING_TYPE:
    if ( !reserve(&user_data->buffer, &user_data->buffer_size, 1+2*strlen(v.value)+1+1) ) return RDF_ERROR;
    outputCSVEscapedString(v.value, user_data->buffer);
    fprintf(file, "LITERAL,%s,", user_data->buffer);
    break;
  case RDF_INT_TYPE:
    fprintf(file, "INTEGER,%i,", v.value);
    break;
  case RDF_FLOAT_TYPE:
    {
      double dval = *((float*)(&v.value));
      fprintf(file, "LITERAL,%f,\"http://www.w3.org/2001/XMLSchema#float\"", dval);
    }
    break;
  case RDF_DATE_TYPE:
  case RDF_BLOB_TYPE:
  default:
    return RDF_ERROR;
  }
  return RDF_OK;
}

/* JvdG */
static RDF_Result CSVSerializer_serializeTriple(RDF_Value source, RDF_Value arc, RDF_Value target, void* serializer_data) {
  CSVSerializerUserData* user_data = serializer_data;
  FILE* file = user_data->file;
  RDF_Result err;
  err = CSVSerializer_serializeValue(source, user_data);
  if ( err != RDF_OK ) return err;
  fputs(",", file);
  err = CSVSerializer_serializeValue(arc, user_data);
  if ( err != RDF_OK ) return err;
  fputs(",", file);
  err = CSVSerializer_serializeValue(target, user_data);
  if ( err != RDF_OK ) return err;
  fputs("\n", file);
  return RDF_OK;
}

