/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors: Ivan Kurmanov, Jasper van de Gronde (win32 port)
 */

#include <string.h>
#include <stdio.h>
#include "../utils/utils.h"
#include <pthread.h>
#ifdef WIN32
#include <windows.h>
#undef INSERT
#undef SELECT
#undef DELETE
#undef CREATE
#undef UPLOAD
#undef UNLOAD
#undef ENTER
#undef LEAVE
#undef DROP
#endif
#include "queryParser.h"
#include <pthread.h>

static QueryParserFunc parserList[12];
/*
int gQueryCounter = 0;
pthread_mutex_t gQueryCounterMutex = PTHREAD_MUTEX_INITIALIZER;

Query gQueryPool[4096];
*/
/* JvdG I doubt this is needed, anywhere.
#ifdef WIN32
#else
#define isspace(x) ((isspace(x) || (x == '\x0D') || (x == '\x0A') || (x == '\r')))
#endif
*/

#define DEBUG_QUERY_PARSER 0

QueryParseResult ParseInsert(Query nq);
QueryParseResult ParseSelect(Query nq);
QueryParseResult ParseDelete(Query nq);
QueryParseResult ParseCreate(Query nq);
QueryParseResult ParseUpload(Query nq);
QueryParseResult ParseUnload(Query nq);
QueryParseResult ParseEnter(Query nq);
QueryParseResult ParseLeave(Query nq);
QueryParseResult ParseDrop(Query nq);
QueryParseResult ParseLogin(Query nq);
QueryParseResult ParseExport(Query nq);


int
initQueryParser() {
  parserList[INSERT] = ParseInsert;
  parserList[SELECT] = ParseSelect;
  parserList[DELETE] = ParseDelete;
  parserList[CREATE] = ParseCreate;
  parserList[UPLOAD] = ParseUpload;
  parserList[UNLOAD] = ParseUnload;
  parserList[ENTER]  = ParseEnter;
  parserList[LEAVE]  = ParseLeave;
  parserList[DROP]   = ParseDrop;
  parserList[LOGIN]  = ParseLogin;
  parserList[EXPORT] = ParseExport;
  return 0;
}

int
deinitQueryParser() {
  return 0;
}

/*
void* parseTermOfType (char* term, RDF_ValueType type) ;
RDF_ValueType termType (const char *term) ;
*/
/*
int
getNextQueryCounter () {
  int ans;
  pthread_mutex_lock(&gQueryCounterMutex);
  ans = gQueryCounter++;
  pthread_mutex_unlock(&gQueryCounterMutex);
  return ans;
}*/

static
QueryType
getQueryType (const char* query) {
  if (startsWith_forgiving_ws(query, "select")) {
    return SELECT;
  } else if (startsWith_forgiving_ws(query, "load")) {
    return UPLOAD;
  } else if (startsWith_forgiving_ws(query, "unload")) {
    return UNLOAD;
  } else if (startsWith_forgiving_ws(query, "enter")) {
    return ENTER;
  } else if (startsWith_forgiving_ws(query, "drop")) {
    return DROP;
  } else if (startsWith_forgiving_ws(query, "leave")) {
    return LEAVE;
  } else if (startsWith_forgiving_ws(query, "insert")) {
    return INSERT;
  } else if (startsWith_forgiving_ws(query, "delete")) {
    return DELETE;
  } else if (startsWith_forgiving_ws(query, "create")) {
    return CREATE;
  } else if (startsWith_forgiving_ws(query, "login")) {
    return LOGIN;
  } else if (startsWith_forgiving_ws(query, "export")) {
    return EXPORT;
  } else return UNKNOWN_QUERY_TYPE;
}

QueryParseResult
parseQuery (const char* queryString, Query *result) {

  QueryType type = getQueryType(queryString);
  Query nq ;
  QueryParseResult err;

  if (!type) {
    return UNKNOWN_QUERY_TYPE_ERROR;
  }

  nq =  (Query)getMem(sizeof(QueryStruct));
  *result = nq;
  bzero(nq, sizeof(QueryStruct));
  nq->type = type;
  nq->queryString = strdup(queryString);

  if (0 != (err = (*(parserList[type]))(nq))) {
    return err;
  }
/* JvdG The query is put into the query pool, but it's never extracted again, seems like a waste to me.
  nq->id = getNextQueryCounter();
  gQueryPool[nq->id] = nq;
*/
  return 0;
}


/***   Next Word   ***/  

/* JvdG Removed support for triple reading as it isn't (or at least shouldn't be) used (anymore). */
static
size_t 
nextWord (const char* str, size_t* start) {
  size_t end;
  /*int literalp;*/

  while ((str[*start] != 0) && (isspace(str[*start]) || (str[*start] == ','))) {
    (*start)++;
  }

  end = *start;

  if (str[end] == 0) return 0;

  /*literalp = (str[end] == '(');*/

  /*while ((str[end] != 0) && (literalp ? (str[end] != ')') : !isspace(str[end]))) {*/
  while ((str[end] != 0) && !isspace(str[end])) {
    end++;
  }

  /*if (str[end] == ')') end++;*/
  return end;
}


/* Jasper van de Gronde
 */
static
Statement*
next_triple( const char * qstring, size_t word_start, size_t* end ) {
  struct triple_pattern_t triple;
  struct state_t state;
  Statement* ans;

  bzero(&triple, sizeof(triple));
  bzero(&state, sizeof(state));

  state.querystart = qstring;
  state.queryend = qstring + strlen(qstring);
  state.curstart = qstring + word_start;
  state.curend = state.curstart;

  if ( !parse_triple_pattern(&triple, &state) ) {
    destroy_triple_pattern(&triple);
    if ( state.errstart == 0 ) {
      return (void*)0;
    }
    else {
      *end = state.errend-state.querystart;
      return (void*)1;
    }
  }

  *end = state.curend - qstring;

  /*  generate the resulting structure */
  
  ans = (Statement*)getMem(sizeof(Statement));
  if ( !ans ) {
    destroy_triple_pattern(&triple);
    return (void*)0;
  }
  bzero(ans, sizeof(Statement));
  ans->source.value   = triple.subject.text;
  ans->source.type    = triple.subject.type;
  ans->arc.value      = triple.predicate.text;
  ans->arc.type       = triple.predicate.type;
  if ( triple.object.type == RDF_INT_TYPE ) {
    ans->target.value = (void*)triple.object.num;
  }
  else {
    ans->target.value = triple.object.text;
  }
  ans->target.type    = triple.object.type;
  ans->optional       = triple.optional;

  /* No need to destroy the triple, as all values that could be freed are now referenced in ans. */

  return ans;
}


static
size_t 
nextVariable (const char* str, size_t* start) {
  size_t end;

  while ((str[*start] != 0) && (isspace(str[*start]) || (str[*start] == ','))) {
    (*start)++;
  }

  end = *start;
  if (str[end] == 0) return 0;
  while ((str[end] != 0) && (!isspace(str[end])) && (str[end] != ',')) {
    end++;
  }
  return end;
}


static
int 
verifyVars (Query nq) {
  Variable v = nq->variables;
  while (v) {
    if (!v->verified && v->outputVariablep) {
      PrintMsg(FAILURE, "Unconstrained variable : ?%s\n", v->name);
      return 1;
    } 
    v = v->next;
  }
  return 0;
}
  
static
void
addLiteralVars (Query nq, RDF_Value val) {
  if (val.type == RDF_VAR_TYPE) {
    Variable v = nq->variables;
    /* JvdG Changed this to add variables in-order */
    if ( v ) {
      for(;;) {
        if (streq(v->name, val.value)) {
          v->verified = 1;
          return;
        }
        if ( v->next ) v = v->next; else break;
      }
      v->next = (Variable)getMem(sizeof(VariableStruct));
      v = v->next;
    }
    else {
      v = (Variable)getMem(sizeof(VariableStruct));
      nq->variables = v;
    }
    v->name = strdup(val.value);
    /*v->next = nq->variables;*/
    if ( nq->showAllVariables ) { /* JvdG Added this bit to allow select * from ... */
      v->outputVariablep = 1;
      v->verified = 1;
    }
  }
}

static
void
addVariableToQuery (Query nq, Variable v) {
  Variable vn = nq->variables;
  Variable prev = vn;

  while (vn) {
    prev = vn;
    vn = vn->next;
  }
  
  if (!prev) {
    nq->variables = v;
  } else {
    prev->next = v;
  }
}


static
QueryParseResult
parse_file_source(Query nq, size_t* start);

/***   Parse Insert   ***/

static
QueryParseResult 
ParseInsert (Query nq) {
  size_t  start = 0;
  size_t  end = 0;
  int insert_mode = 0; /* JvdG Added to support file insert queries. */

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "insert", end-start)) {
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
    return GENERAL_SYNTAX_ERROR;
  }
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "into", end-start)==0) {
    /* Normal insert query, a database name and a list of triples follow. */
  }
  else {
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Missing into.", start, end);
    return GENERAL_SYNTAX_ERROR;
  }
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "from", end-start)==0) {
    insert_mode = 2;
    start = end;
  }

  switch(insert_mode) {
  case 0:
    /* Normal insert query: insert into DB TRIPLES. */
    for(;;) {
      Statement* l;
    
#if DEBUG_QUERY_PARSER
      PrintMsg(DEBUG_INFO, "before next_triple(), qstring: %s\n", &nq->queryString[start] );
#endif

      l = next_triple( nq->queryString, start, &end );

#if DEBUG_QUERY_PARSER
      PrintMsg(DEBUG_INFO, "after next_triple(), end-start: %i ; end: %i\n", end-start, end );
#endif

      if (l == (void*)1) {
        snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
        return GENERAL_SYNTAX_ERROR;
      }
      if (l && end-start > 0) {
        l->next = nq->statements;
        nq->statements = l;
        start = end;
      } else {
        delete_statement(l);
        break;
      }
    }
    break;
  case 1:
    /* Where insert query: insert into DB TRIPLES where WHERE_CLAUSE. */
    /** \todo Implement support for "where" insert queries. */
    break;
  case 2: /* JvdG Added to support file insert queries. */
    /* File insert query: insert into DB from FORMAT <FILE>. */
    {
      int err;

      nq -> type = INSERTFILE;
    
      err = parse_file_source(nq, &start);
      if ( err ) {
        return err;
      }
    }
    break;
  }
  
  return NO_PARSE_ERROR;
}

int parse_where_statement ( Query nq, size_t end, size_t start );

/***   Parse Select   ***/

/* JvdG Made this accept select * from ... */
static
QueryParseResult 
ParseSelect (Query nq) {

  size_t  start = 0;
  size_t  end = 0;
  int err;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "select", end-start)) {
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Missing select!", start, end);
    return GENERAL_SYNTAX_ERROR;
  }
  start = end;

  while (end) {
    Variable v;
    end = nextVariable(nq->queryString, &start);
    if (end == 0) {
      snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, start);
      return GENERAL_SYNTAX_ERROR;
    }
    if (strncasecmp(&nq->queryString[start], "from", end-start) == 0) {
      start = end;
      break;
    }
    else if ( nq->showAllVariables ) {
      snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Missing from!", start, end);
      return GENERAL_SYNTAX_ERROR;
    }
    else if (nq->queryString[start] == '?') {
      ++start; /* move past the '?' */
    
      v = (Variable)getMem(sizeof(VariableStruct));
      bzero(v, sizeof(VariableStruct));
      v->name = (char*)getMem(end-start+1);
      bzero(v->name, end-start+1);    
      memcpy(v->name, &nq->queryString[start], end-start);
      addVariableToQuery(nq, v);
      v->outputVariablep = 1;
      start = end;
    }
    else if ( !nq->variables && end-start == 1 && nq->queryString[start] == '*' ) {
      nq->showAllVariables = TRUE;
      start = end;
    }
    else {
      snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
      return GENERAL_SYNTAX_ERROR;
    }
  }

 
  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "where", end-start)) {
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Missing where!", start, end);
    return GENERAL_SYNTAX_ERROR;
  }
  start = end;

  
  nq->outputType = TAB_LIMITED_OUTPUT;


  err = parse_where_statement( nq, end, start );
  return err ? err : NO_PARSE_ERROR;
}


/***   Parse Delete   ***/

static
QueryParseResult 
ParseDelete(Query nq) {
  size_t  start = 0;
  size_t  end = 0;
  int delete_mode =0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "delete", end-start)) {
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
    return GENERAL_SYNTAX_ERROR;
  }
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "from", end-start)) {
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Missing from.", start, end);
    return GENERAL_SYNTAX_ERROR;
  }
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!strncasecmp(&nq->queryString[start], "where", end-start)) { 
    delete_mode = 1; 
    start = end;
  }
  else if (!strncasecmp(&nq->queryString[start], "from", end-start)) { /* JvdG Added to support file delete queries. */
    delete_mode = 2; 
    start = end;
  }

  switch(delete_mode) {
  case 0:
    /* Normal delete query: delete from DB TRIPLES */
    while (end) {
      Statement* l;
    
      PrintMsg(DEBUG_INFO, "before next_triple(), qstring: %s\n", &nq->queryString[start] );
      l = next_triple( nq->queryString, start, &end );
      PrintMsg(DEBUG_INFO, "after next_triple(), end-start: %i ; end: %i\n", end-start, end );
      
      if (l == (void*)1) {
        snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
        return GENERAL_SYNTAX_ERROR;
      }
      if (l && end-start > 0) {
        /* parseLiteral(&nq->queryString[start], end-start);*/
        l->next = nq->statements;
        nq->statements = l;
        start = end;
      } else {
        delete_statement(l);
        break;
      }
    }
    break;
  case 1:
    /* Where delete query: delete from DB where TRIPLES */
    {
      int err;
      Variable var ;

      nq -> type = DELETEWHERE;
      nq -> outputType = 0;

      err = parse_where_statement( nq, end, start );
      if( err ) {
        return err;
      }

      var = nq->variables;
      while (var) {
        var -> outputVariablep = 1;
        var = var->next;
      }
    }
    break;
  case 2: /* JvdG Added to support file delete queries. */
    /* File delete query: delete from DB from FORMAT <FILE> */
    {
      int err;

      nq -> type = DELETEFILE;
    
      err = parse_file_source(nq, &start);
      if ( err ) {
        return err;
      }
    }
    break;
  }
 
    /****
    ****/
  
  return NO_PARSE_ERROR;
}


static
QueryParseResult 
ParseCreate (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "create", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "database", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;
  
  return NO_PARSE_ERROR;
}

static
QueryParseResult 
ParseDrop (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "drop", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "database", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;
  
  return NO_PARSE_ERROR;
}

static
QueryParseResult 
ParseEnter (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "enter", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "namespace", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!end) return GENERAL_SYNTAX_ERROR;
  if (strncasecmp(&nq->queryString[start], "xmlns", 5)) return GENERAL_SYNTAX_ERROR;
  nq->namespacePrefix = (char*)getMem(end-start+1);
  bzero(nq->namespacePrefix, end-start+1);
  memcpy(nq->namespacePrefix, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!end) return GENERAL_SYNTAX_ERROR;
  nq->namespace = (char*)getMem(end-start+1);
  bzero(nq->namespace, end-start+1);
  memcpy(nq->namespace, &nq->queryString[start], end-start);
  start = end;
  
  return NO_PARSE_ERROR;
}

static
QueryParseResult 
ParseLeave (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "leave", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "namespace", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!end) return GENERAL_SYNTAX_ERROR;
  if (strncasecmp(&nq->queryString[start], "xmlns", 5)) return GENERAL_SYNTAX_ERROR;
  nq->namespacePrefix = (char*)getMem(end-start+1);
  bzero(nq->namespacePrefix, end-start+1);
  memcpy(nq->namespacePrefix, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (end)   return GENERAL_SYNTAX_ERROR;
  return NO_PARSE_ERROR;
  
}

static
QueryParseResult 
ParseUpload (Query nq) {
  size_t  start = 0;
  size_t  end = 0;
  int err;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "load", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  err = parse_file_source(nq, &start);
  if ( err ) {
    return err;
  }

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "into", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (end != 0) return GENERAL_SYNTAX_ERROR;

  return NO_PARSE_ERROR;
}

/* JvdG Moved this into a separate function so it can also be used in ParseInsert. */
static
QueryParseResult
parse_file_source(Query nq, size_t* start_p) {
  size_t start = *start_p;
  size_t end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "rss", end-start) == 0) {
    nq->fileType = RSS;
  } else if (strncasecmp(&nq->queryString[start], "rdf_xml", end-start) == 0) {
    nq->fileType = RDF_XML;
  } else if (strncasecmp(&nq->queryString[start], "ntriples", end-start) == 0) {
    nq->fileType = NTRIPLES;
  } else if (strncasecmp(&nq->queryString[start], "turtle", end-start) == 0) {
    nq->fileType = TURTLE;
  } else if (strncasecmp(&nq->queryString[start], "csv", end-start) == 0) {
    nq->fileType = CSV;
  /*} else if (strncasecmp(&nq->queryString[start], "dmoz_rdf", end-start) == 0) {
    nq->fileType = DMOZ_RDF_XML;*/
  } else {
    return RDF_UNKNOWN_FILE_FORMAT_ERROR;
  }
  start = end;

  /* Jasper van de Gronde
     The if below is commented out because it should be able to load files.
     Also, if this word is read here the next word will be "into" instead of the url.
  */
  /*
  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "file", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;
  */

  /* JvdG made this use parse_uri */
  {
    struct state_t state;
    bzero(&state, sizeof(state));
    state.querystart = nq->queryString;
    state.queryend = nq->queryString + strlen(nq->queryString);
    state.curstart = state.querystart + start;
    state.curend = state.curstart;
    nq->url = parse_uri(&state);
    end = state.curend - state.querystart;
    start = end;
  }
  /*
  end = nextWord(nq->queryString, &start);
  nq->url = (char*) getMem(end-start+1);
  memcpy(nq->url, &nq->queryString[start], end-start);
  start = end;
  */

  *start_p = start;

  return NO_PARSE_ERROR;
}


static
QueryParseResult 
ParseUnload (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "unload", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "file", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  /* JvdG made this use parse_uri */
  {
    struct state_t state;
    bzero(&state, sizeof(state));
    state.querystart = nq->queryString;
    state.queryend = nq->queryString + strlen(nq->queryString);
    state.curstart = state.querystart + start;
    state.curend = state.curstart;
    nq->url = parse_uri(&state);
    end = state.curend - state.querystart;
    start = end;
  }
  /*
  end = nextWord(nq->queryString, &start);
  nq->url = (char*) getMem(end-start+1);
  memcpy(nq->url, &nq->queryString[start], end-start);
  start = end;
  */

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "from", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (end != 0) return GENERAL_SYNTAX_ERROR;

  return NO_PARSE_ERROR;

}

/* JvdG
   Moved a lot of if's out of the while loop.
   Added/modified error reporting.
   Added the non-null option.
*/
static int parse_where_statement ( Query nq, size_t end, size_t start ) {
  Statement* curstatement = 0;

  while (end) {
    Statement* l;
    
#if DEBUG_QUERY_PARSER
    PrintMsg(DEBUG_INFO, "before next_triple(), qstring: %s\n", &nq->queryString[start] );
#endif

    l = next_triple( nq->queryString, start, &end );

#if DEBUG_QUERY_PARSER
     PrintMsg(DEBUG_INFO, "after next_triple(), end-start: %i ; end: %i, returned: %i\n", 
             end-start, 
             end,
             (long) l );
#endif
    
    if ( l == (void*)0 ) {
      /* Couldn't read a triple, but it wasn't a syntax error */
      break;
    } else if( l== (void*)1 ) {
      /* Syntax error */
      snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
      return GENERAL_SYNTAX_ERROR;
    } else if (end-start > 0) {
      /* parseLiteral(&nq->queryString[start], end-start); */

      addLiteralVars(nq, l->source);
      addLiteralVars(nq, l->arc);
      /*if (l->target.type == RDF_RESOURCE_TYPE)*/ addLiteralVars(nq, l->target);

      /* JvdG modified this to create them in order */
      if ( curstatement ) {
        curstatement->next = l;
      }
      else {
        nq->statements = l;
      }
      curstatement = l;
      /*
      l->next = nq->literals;
      nq->literals = l;
      */
      start = end;

    } else {
      delete_statement(l);
      /* Shouldn't happen? */
      return -4; /* -4 already has a very unclear meaning, but perhaps it would be better to use something else */
    }
  }

  /* Check for AND */
  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "and", end-start) == 0) {
    /* Jasper van de Gronde
     * Parse the constraint list.
     */

#if DEBUG_QUERY_PARSER
    PrintMsg(DEBUG_INFO, "'and' found\n" );
#endif

    start = end;

    {
      struct state_t state;
      bzero(&state, sizeof(state));
      state.querystart = nq->queryString;
      state.queryend = nq->queryString + strlen(nq->queryString);
      state.curstart = nq->queryString + start;
      state.curend = state.curstart;
      nq->constraint_list = parse_constraint_list(&state);
      start = state.curend - nq->queryString;
      if ( state.errstart ) {
        start = state.errstart - nq->queryString;
        end = state.errend - nq->queryString;
        snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
        return GENERAL_SYNTAX_ERROR;
      }
    }
  }
  
  /* Check for USING */
  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "using", end-start) == 0) {
    /* Jasper van de Gronde
     * Parse the namespace list.
     */

#if DEBUG_QUERY_PARSER
    PrintMsg(DEBUG_INFO, "'using' found\n" );
#endif

    start = end;

    {
      struct state_t state;
      bzero(&state, sizeof(state));
      state.querystart = nq->queryString;
      state.queryend = nq->queryString + strlen(nq->queryString);
      state.curstart = nq->queryString + start;
      state.curend = state.curstart;
      nq->for_list = parse_for_list(&state);
      start = state.curend - nq->queryString;
      if ( state.errstart ) {
        start = state.errstart - nq->queryString;
        end = state.errend - nq->queryString;
        snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i", start, end);
        return GENERAL_SYNTAX_ERROR;
      }
    }
  }

  /* Check for "non-null */
  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "non-null", end-start) == 0) {
#if DEBUG_QUERY_PARSER
    PrintMsg(DEBUG_INFO, "'non-null' found\n" );
#endif

    /*nq->allow_null = 0;*/
    PrintMsg(INFO, "\n***** non-null is deprecated, use ?(src arc trg) instead!\n\n");

    start = end;
  }
  else {
    /*nq->allow_null = 1;*/
  }

  /* Check for output format */
  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "output", end-start) == 0) {
#if DEBUG_QUERY_PARSER
    PrintMsg(DEBUG_INFO, "'output' found\n" );
#endif

    start = end;
    end = nextWord(nq->queryString, &start);
    if (strncasecmp(&nq->queryString[start], "rdf-xml", end-start) == 0) {
      nq->outputType = RDF_XML_OUTPUT;
    }
    else if (strncasecmp(&nq->queryString[start], "javascript", end-start) == 0) {
      nq->outputType = JAVASCRIPT_OUTPUT;
    }
    else if (strncasecmp(&nq->queryString[start], "tab-delimited", end-start) == 0) {
      nq->outputType = TAB_LIMITED_OUTPUT;
    }
    else if (strncasecmp(&nq->queryString[start], "variable-list", end-start) == 0) {
      nq->outputType = VAR_LIST_OUTPUT;
    }
    else {
      snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Unrecognized output format!", start, end);
      return GENERAL_SYNTAX_ERROR;
    }
    start = end;
  }

  end = nextWord(nq->queryString, &start);
  if ( end != 0 ) {
    /* Never reached the end */
    snprintf(nq->error_message, QUERY_ERR_LEN, "%i;%i:Unrecognized end.", start, strlen(nq->queryString));
    return GENERAL_SYNTAX_ERROR;
  }
  
#if DEBUG_QUERY_PARSER
  PrintMsg(DEBUG_INFO, "at the end of parse_where_statement()\n" );
#endif

  if (!nq->statements) return MALFORMED_LITERAL_ERROR;
  if (verifyVars(nq)) return RDF_UNCONSTRAINED_VARIABLE;

#if DEBUG_QUERY_PARSER
  PrintMsg(DEBUG_INFO, "after verifyVars(nq) in parse_where_statement(), before exit\n" );
#endif

  return NO_PARSE_ERROR;

}

/* JvdG */
static QueryParseResult ParseLogin(Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  bzero(&nq->identity, sizeof(identity_t));

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "login", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  {
    struct state_t state;
    bzero(&state, sizeof(state));
    state.curstart = nq->queryString + start;
    state.curend = state.curstart;
    state.querystart = nq->queryString;
    state.queryend = nq->queryString + strlen(nq->queryString);
    
    nq->identity.username = parse_text_literal(&state);
    if ( !nq->identity.username || !*(nq->identity.username) ) {
      if ( nq->identity.username ) {
        free((void*)nq->identity.username);
        nq->identity.username = 0;
      }
      return GENERAL_SYNTAX_ERROR;
    }

    nq->identity.password = parse_text_literal(&state);
    if ( !nq->identity.password ) { /* Password can be empty. */
      if ( nq->identity.username ) {
        free((void*)nq->identity.username);
        nq->identity.username = 0;
      }
      return GENERAL_SYNTAX_ERROR;
    }

    while(state.curend != state.queryend && isspace(*state.curend) ) ++state.curend;
    if ( state.curend != state.queryend ) {
      if ( nq->identity.username ) {
        free((void*)nq->identity.username);
        nq->identity.username = 0;
      }
      if ( nq->identity.password ) {
        free((void*)nq->identity.password);
        nq->identity.password = 0;
      }
      return GENERAL_SYNTAX_ERROR;
    }
  }
/*
  end = nextWord(nq->queryString, &start);
  if ( end == 0 ) return GENERAL_SYNTAX_ERROR;
  nq->identity.username = strduprange(&nq->queryString[start], &nq->queryString[end]);
  start = end;

  end = nextWord(nq->queryString, &start);
  if ( end == 0 ) {
    nq->identity.password = strduprange(0, 0);
  }
  else {
    nq->identity.password = strduprange(&nq->queryString[start], &nq->queryString[end]);
  }
  start = end;
*/
  return NO_PARSE_ERROR;
}

/* JvdG */
static QueryParseResult ParseExport(Query nq) {
  size_t  start = 0;
  size_t  end = 0;
  int err;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "export", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "raw", end-start) == 0) {
    nq->exportRaw = 1;
    start = end;
    end = nextWord(nq->queryString, &start);
  }
  else {
    nq->exportRaw = 0;
  }
  if (strncasecmp(&nq->queryString[start], "to", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  err = parse_file_source(nq, &start);
  if ( err ) {
    return err;
  }

  end = nextWord(nq->queryString, &start);
  if (end != 0) return GENERAL_SYNTAX_ERROR;

  return NO_PARSE_ERROR;
}
