/*
 * Example of server using TCP protocol.
 *
 * Ported to Windows by Jasper van de Gronde.
 */


#include	<stdio.h>
#include	<sys/types.h>
#include    "../query/queryParser.h" /* Added */
#ifdef WIN32
#include    <winsock2.h>
#else
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#endif
#include        "../utils/utils.h"
#include        "connection.h"
#include "security.h"


#define MAXLINE 512
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#define QUERY_END "</>"
#define RESPONSE_END  NEWLINE "</>" NEWLINE

static unsigned short server_port = 0;

static SOCKET gListenSockfd = 0;

int ProcessQuery (const char* query,  ClientConnection cl);
static int CloseListenSocket();
static unsigned short getServerPort();

/* JvdG */
int initServer() {
  int failures = 0;
#ifdef WIN32
  struct WSAData wsa_data;
  if ( 0 != WSAStartup(MAKEWORD( 2, 2 ), &wsa_data) ) ++failures;
#endif
  if ( 0 != init_connection() ) ++failures;
  if ( !(server_port=getServerPort()) ) ++failures;
  return failures;
}

/* JvdG */
int deinitServer() {
  int failures = 0;
  if ( 0 != CloseListenSocket() ) ++failures;
  if ( 0 != deinit_connection() ) ++failures;
#ifdef WIN32
  if ( 0 != WSACleanup() ) ++failures;
#endif
  return failures;
}

static int CloseListenSocket() {
  if ( gListenSockfd ) {
    closesocket(gListenSockfd);
    gListenSockfd = 0;
  }
  return 0;
}

/* JvdG Made this use an unsigned short */
static
unsigned short
getServerPort () {
  const char *port = get_setting("PORT");
  if (port) {
    char* endchar;
    unsigned short ip = (unsigned short)strtoul(port, &endchar, 10);
    if (ip) {
      PrintMsg(INFO, "Binding = %i\n", ip);
      return ip;
    } else {
      return 7001;
    }
  } else {
    PrintMsg(CRITICAL, "\nPlease set the server port by setting the environment variable RDFDB_PORT\n.e.g., export RDFDB_PORT=7001\n");
    return 0;
  }  
}

#ifdef WIN32
#else
static
int 
catchBrokenPipe (int sig) {
  ClientConnection conn =  pthread_getspecific(gCCKey);
  conn->socket = 0;
  PrintMsg(DEBUG_INFO, "Catching broken pipe signal\n");
  signal(SIGPIPE, catchBrokenPipe);
  return 0;
}
#endif

void
addNamespace (ClientConnection cl, char *prefix, char* url) {
  Namespace ns = (Namespace) getMem(sizeof(NamespaceStruct));
  ns->prefix = strdup(prefix);
  ns->url    = strdup(url);
  ns->next = cl->namespaces;
  cl->namespaces = ns;
}

void
freeNamespace (Namespace ns) {
  freeMem(ns->prefix);
  freeMem(ns->url);
  freeMem((char*)ns);
}

void removeNamespace (ClientConnection cl, char* prefix) {
  Namespace next = cl->namespaces;
  Namespace prev = next;
  while (next) {
    if (streq(prefix, next->prefix)) {
      if (prev == next) {
	      cl->namespaces = next->next;
      } else {
	      prev->next = next->next;
      }
      freeNamespace(next);
      break;
    }
    prev = next;
    next = next->next;
  }
}

static
int closeAndExit (int sig) {
  /* Jasper van de Gronde
   * Added SIGTERM and moved all the functionality to a different function.
   * Changed this to make use of the new state machine, which means this just causes openService to end.
   */
  if (sig == SIGINT) {
    PrintMsg(INFO, "Exiting on SIGINT ...\n");
    CloseListenSocket(); /* JvdG This relies on the accept call aborting when the socket is closed. On Win32 this seems to work. */
  }
  else if (sig == SIGTERM) {
    PrintMsg(INFO, "Exiting on SIGTERM ...\n");
    CloseListenSocket(); /* JvdG This relies on the accept call aborting when the socket is closed. On Win32 this seems to work. */
  }
  return 0;
}

/* JvdG Modified this to make use of the new client connection functions. */
int
openService() {
  SOCKET newsocket;
  int clilen;
  struct sockaddr_in	cli_addr, serv_addr;
  /*pthread_t thread;*/
  /*int err;*/
  /*pthread_attr_t attr;*/

  PrintMsg(INFO, "Server starting ...\n");

  /* Open a TCP socket */
  
#ifdef WIN32
  if ((gListenSockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
#else
  if ((gListenSockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
#endif
    PrintMsg(CRITICAL, "server: can't open stream socket");
    return -1;
  }
  
  /*  Bind our local address so that the client can send to us. */
  
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(server_port);
  
  PrintMsg(INFO, "Binding local socket ...");
  while (bind(gListenSockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
    sleep(1);
  }
  PrintMsg(INFO, "done\n");

  signal(SIGINT, closeAndExit);  
  signal(SIGTERM, closeAndExit);  

  listen(gListenSockfd, 128);

  /* JvdG Not used?: pthread_key_create(&gCCKey, 0); */

  /*pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);*/

  PrintMsg(INFO, "Awaiting connections ...\n"); /* JvdG Put this outside the loop and changed the level to INFO instead of DEBUG_INFO. */

  for ( ; ; ) {
    /* Wait for a connection from a client process. */    
    clilen = sizeof(cli_addr);
    newsocket = accept(gListenSockfd, (struct sockaddr *) &cli_addr,  &clilen);
#ifdef WIN32
    if ( newsocket == INVALID_SOCKET ) {
#else
    if (newsocket < 0) {
#endif
      PrintMsg(CRITICAL, "server: accept error\n");
      break;
    } else if ( clilen == sizeof(struct sockaddr_in) && cli_addr.sin_family == AF_INET ) {      
      const unsigned int client_ipv4 = (cli_addr.sin_addr.S_un.S_un_b.s_b1 << 24) | (cli_addr.sin_addr.S_un.S_un_b.s_b2 << 16) | (cli_addr.sin_addr.S_un.S_un_b.s_b3 << 8) | (cli_addr.sin_addr.S_un.S_un_b.s_b4 << 0);
      if ( 0 == new_client_connection(newsocket, client_ipv4) ) {
        PrintMsg(CRITICAL, "Failed to create connection.\n");
      }
    }
  }
  /*pthread_attr_destroy(&attr);*/

  PrintMsg(INFO, "Server stopped.\n");

  return 0;
}
